'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _debug = require('../debug');

var _debug2 = _interopRequireDefault(_debug);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

class Scope {
  constructor(options) {
    const value = options.value,
          parent = options.parent,
          sourceId = options.sourceId,
          sources = options.sources,
          refs = options.refs,
          start = options.start,
          path = options.path;
    // A single source cahce shared between all scopes in the same
    // execution chain.

    this.sources = sources || {};
    // The source this value came from (used when joins are encountered
    // to get at the entire collection)
    this.sourceId = sourceId;
    // The parent scope of this scope
    this.parent = parent;
    // The specific value of this scope
    this.value = value;
    // The path of this specific value from source
    this.path = path || [];
    // A cache of all refs seen in all documents fetched so far, shared between all scopes in
    // the same execution chain
    this.refs = refs || {};
    // Start offset of an array, used to represent partial sets, like e.g. the response
    // to *[100..102] will most likely be a two item array with start == 100 if the fetcher
    // is able to only fetch the relevant subset of the collection
    this.start = start;
  }

  claimSource(id) {
    const result = {};
    this.sources[id] = new Promise((resolve, reject) => {
      result.resolve = resolve;
      result.reject = reject;
    });
    return result;
  }

  cacheRefs(refs) {
    if (!refs) return;
    Object.keys(refs).forEach(id => {
      this.refs[id] = refs[id];
    });
  }

  slice(start, end) {
    if (!Array.isArray(this.value)) {
      return this.clone({
        value: null
      });
    }
    const collectionStart = this.start || 0;
    return this.clone({
      path: this.path.concat({
        op: 'subscript',
        start: start,
        end: end,
        first: false
      }),
      value: this.value.slice(start + collectionStart, end + collectionStart),
      start: 0
    });
  }

  first(offset) {
    if (!Array.isArray(this.value)) {
      return this.clone({
        value: null
      });
    }
    const collectionStart = this.start || 0;
    return this.clone({
      path: this.path.concat({
        op: 'subscript',
        start: offset,
        end: offset + 1,
        first: true
      }),
      value: this.value.slice(offset + collectionStart)[0],
      start: 0
    });
  }

  // Returns true if the provided document references any of the ids
  doesReference(ids) {
    ids = Array.isArray(ids) ? ids : [ids];

    const document = this.value;

    (0, _debug2.default)('doesReference()', document, ids);

    if (!document || !typeof document === 'object' || !document._id) {
      return false;
    }

    let refs = this.refs[document._id];
    if (!refs) {
      refs = mapReferences(document);
      (0, _debug2.default)('mapReferences()', refs, document);
      this.cacheRefs({
        [document._id]: refs
      });
    }

    return !!refs.find(ref => ids.indexOf(ref.id) !== -1);
  }

  sourceIsCached(sourceId) {
    return sourceId in this.sources;
  }

  dataForSource(sourceId) {
    return this.sources[sourceId];
  }

  clone(options) {
    return new Scope(Object.assign({
      sources: this.sources,
      parent: this.parent,
      value: this.value,
      path: this.path,
      sourceId: this.sourceId,
      refs: this.refs,
      start: this.start
    }, options || {}));
  }

  child(options) {
    return new Scope(Object.assign({
      parent: this,
      sources: this.sources,
      sourceId: this.sourceId,
      refs: this.refs
    }, options || {}));
  }

  resolveAccessor(path) {
    (0, _debug2.default)('resolve accessor:', this.value, path);
    let scope = this;
    if (scope.value) {
      for (let i = 0; i < path.length; i++) {
        const operation = path[i];
        switch (operation.op) {
          case 'parent':
            scope = scope.parent;
            (0, _debug2.default)('^ =>', scope.value);
            break;
          case 'attribute':
            scope = scope.child({
              parent: scope,
              value: scope.value[operation.name],
              path: scope.path.concat(operation)
            });
            (0, _debug2.default)(operation.name, '=>', scope.value, 'path:', scope.path);
            break;
          default:
            throw new Error(`Unkown accessor path element ${operation.op}`);
        }
        if (!scope.value) {
          break;
        }
      }
    }

    // Unresolved attributes should return null, not undefined?
    if (scope.value === undefined) {
      scope = scope.clone({
        value: null
      });
    }
    return scope;
  }

  // Used in joins to find every possible value for this scope in every other document
  // from the same source
  allSiblings() {
    var _this = this;

    return _asyncToGenerator(function* () {
      (0, _debug2.default)("sourceId", _this.sourceId, "sources", _this.sources);
      const source = yield _this.dataForSource(_this.sourceId);
      if (!source) {
        return [];
      }
      return source.documents.map(function (item) {
        const scope = _this.clone({
          path: [],
          value: item
        });
        (0, _debug2.default)('resolving', scope.value, _this.pathAsString());
        return scope.resolveAccessor(_this.path);
      }).filter(function (scope) {
        return !!scope.value;
      });
    })();
  }

  // Resolves an accessor for all possible values of that accessor. Used
  // to expand join expressions so that i.e. ^.foo._ref resolves to all
  // foo._refs of the parent expression. Always returns an array of scopes
  resolveAccessorForAll(path) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      (0, _debug2.default)('resolve accessor (all):', _this2.value, path);
      let scope = _this2;
      for (let i = 0; i < path.length; i++) {
        const operation = path[i];
        switch (operation.op) {
          case 'parent':
            scope = scope.parent;
            (0, _debug2.default)('^ =>', scope.value);
            break;
          case 'attribute':
            if (scope.sourceId) {
              const source = (yield scope.dataForSource(scope.sourceId)).documents;
              const subPath = path.slice(i);
              const result = [];
              source.forEach(function (item) {
                const itemScope = scope.clone({
                  value: item
                });
                const resolvedScope = itemScope.resolveAccessor(subPath);
                if (resolvedScope.value !== null) {
                  result.push(resolvedScope);
                }
              });
              return result;
            }

            if (scope.value === null || scope.value === undefined) {
              return scope.clone({
                value: null
              });
            }

            scope = scope.child({
              parent: scope,
              value: scope.value[operation.name]
            });
            (0, _debug2.default)(operation.name, '=>', scope.value);
            break;
          default:
            throw new Error(`Unkown accessor path element ${operation.op}`);
        }
        if (!scope.value) {
          break;
        }
      }
      // Unresolved attributes should return null, not undefined?
      if (scope.value === undefined) {
        scope = scope.clone({
          value: null
        });
      }
      return [scope];
    })();
  }

  pathAsString() {
    return this.path.map(element => {
      switch (element.op) {
        case 'attribute':
          return element.name;
        case 'subscript':
          if (element.first) {
            return `[${element.start}]`;
          }
          return `[${element.start}...${element.end}]`;
        default:
          throw new Error(`Unknown path op ${element.op}`);
      }
    }).join('.');
  }

  inspect() {
    return `Scope<${this.pathAsString()}: ${JSON.stringify(this.value)}>`;
  }

}
exports.default = Scope; // A small function to extract all refs in a document as an array

function mapReferences(value) {
  (0, _debug2.default)('mapping', value);
  if (value === null || value === undefined) return [];
  if (typeof value === 'object') {
    if (value._ref) {
      return [{
        id: value._ref
      }];
    }
    return Object.keys(value).reduce((result, key) => result.concat(mapReferences(value[key])), []);
  }
  if (Array.isArray(value)) {
    return value.reduce((result, item) => result.concat(mapReferences(item)), []);
  }
  return [];
}
//# sourceMappingURL=Scope.js.map